@@ -1,11 +1,11 @@
 ;;; easy-kill.el --- kill & mark things easily       -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2013-2014  Free Software Foundation, Inc.
+;; Copyright (C) 2013-2022  Free Software Foundation, Inc.
 
 ;; Author: Leo Liu <sdl.web@gmail.com>
-;; Version: 0.9.4
+;; Version: 0.9.5
 ;; Package-Type: simple
-;; Package-Requires: ((emacs "24") (cl-lib "0.5"))
+;; Package-Requires: ((emacs "25") (cl-lib "0.5"))
 ;; Keywords: killing, convenience
 ;; Created: 2013-08-12
 ;; URL: https://github.com/leoliu/easy-kill
@@ -27,13 +27,13 @@
 
 ;; `easy-kill' aims to be a drop-in replacement for `kill-ring-save'.
 ;;
-;; To use: (global-set-key [remap kill-ring-save] 'easy-kill)
+;; To use: (global-set-key [remap kill-ring-save] #'easy-kill)
 
 ;; `easy-mark' is similar to `easy-kill' but marks the region
 ;; immediately. It can be a handy replacement for `mark-sexp' allowing
 ;; `+'/`-' to do list-wise expanding/shrinking.
 ;;
-;; To use: (global-set-key [remap mark-sexp] 'easy-mark)
+;; To use: (global-set-key [remap mark-sexp] #'easy-mark)
 
 ;; Please send bug reports or feature requests to:
 ;;      https://github.com/leoliu/easy-kill/issues
@@ -42,64 +42,39 @@
 
 (require 'cl-lib)
 (require 'thingatpt)
-(eval-when-compile (require 'cl))       ;For `defsetf'.
 
-(eval-and-compile
-  (cond
-   ((fboundp 'set-transient-map) nil)
-   ((fboundp 'set-temporary-overlay-map) ; new in 24.3
-    (defalias 'set-transient-map 'set-temporary-overlay-map))
-   (t
-    (defun set-transient-map (map &optional keep-pred)
-      (let* ((clearfunsym (make-symbol "clear-temporary-overlay-map"))
-             (overlaysym (make-symbol "t"))
-             (alist (list (cons overlaysym map)))
-             (clearfun
-              `(lambda ()
-                 (unless ,(cond ((null keep-pred) nil)
-                                ((eq t keep-pred)
-                                 `(eq this-command
-                                      (lookup-key ',map
-                                                  (this-command-keys-vector))))
-                                (t `(funcall ',keep-pred)))
-                   (set ',overlaysym nil) ;Just in case.
-                   (remove-hook 'pre-command-hook ',clearfunsym)
-                   (setq emulation-mode-map-alists
-                         (delq ',alist emulation-mode-map-alists))))))
-        (set overlaysym overlaysym)
-        (fset clearfunsym clearfun)
-        (add-hook 'pre-command-hook clearfunsym)
-        (push alist emulation-mode-map-alists))))))
-
-(defcustom easy-kill-alist '((?w word     " ")
-                             (?s sexp     "\n")
-                             (?l list     "\n")
-                             (?f filename "\n")
-                             (?d defun    "\n\n")
-                             (?e line     "\n")
+(defcustom easy-kill-alist '((?w word           " ")
+                             (?s sexp           "\n")
+                             (?l list           "\n")
+                             (?f filename       "\n")
+                             (?d defun          "\n\n")
+                             (?D defun-name     " ")
+                             (?e line           "\n")
                              (?b buffer-file-name))
   "A list of (CHAR THING APPEND).
 CHAR is used immediately following `easy-kill' to select THING.
 APPEND is optional and if non-nil specifies the separator (a
-string) for appending current selection to previous kill.
-
-Note: each element can also be (CHAR . THING) but this is
-deprecated."
+string) for appending current selection to previous kill."
   :type '(repeat (list character symbol
                        (choice string (const :tag "None" nil))))
   :group 'killing)
 
+(defcustom easy-kill-cycle-ignored nil
+  "A list of things that `easy-kill-cycle' should ignore."
+  :type '(repeat symbol)
+  :group 'killing)
+
 (defcustom easy-kill-unhighlight-key nil
   "Key sequence if non-nil to unhighlight the kill candidate."
   :type '(choice (const :tag "None" nil) key-sequence)
   :group 'killing)
 
-(defcustom easy-kill-try-things '(url email line)
+(defcustom easy-kill-try-things '(url email uuid line)
   "A list of things for `easy-kill' to try."
   :type '(repeat symbol)
   :group 'killing)
 
-(defcustom easy-mark-try-things '(url email sexp)
+(defcustom easy-mark-try-things '(url email uuid sexp)
   "A list of things for `easy-mark' to try."
   :type '(repeat symbol)
   :group 'killing)
@@ -114,30 +89,39 @@ deprecated."
 
 (defvar easy-kill-base-map
   (let ((map (make-sparse-keymap)))
-    (define-key map "-" 'easy-kill-shrink)
-    (define-key map "+" 'easy-kill-expand)
-    (define-key map "=" 'easy-kill-expand)
-    (define-key map "@" 'easy-kill-append)
+    (define-key map "-" #'easy-kill-shrink)
+    (define-key map "+" #'easy-kill-expand)
+    (define-key map "=" #'easy-kill-expand)
+    (define-key map " " #'easy-kill-cycle)
+    (define-key map "@" #'easy-kill-append)
     ;; Note: didn't pick C-h because it is a very useful prefix key.
-    (define-key map "?" 'easy-kill-help)
-    (define-key map [remap set-mark-command] 'easy-kill-mark-region)
-    (define-key map [remap kill-region] 'easy-kill-region)
-    (define-key map [remap keyboard-quit] 'easy-kill-abort)
+    (define-key map "?" #'easy-kill-help)
+    (define-key map [remap set-mark-command] #'easy-kill-mark-region)
+    (define-key map [remap kill-region] #'easy-kill-region)
+    (define-key map [remap delete-region] #'easy-kill-delete-region)
+    (define-key map [remap keyboard-quit] #'easy-kill-abort)
     (define-key map [remap exchange-point-and-mark]
-      'easy-kill-exchange-point-and-mark)
+      #'easy-kill-exchange-point-and-mark)
     (mapc (lambda (d)
-            (define-key map (number-to-string d) 'easy-kill-digit-argument))
+            (define-key map (number-to-string d) #'easy-kill-digit-argument))
           (number-sequence 0 9))
     map))
 
 (defvar easy-kill-inhibit-message nil)
 
+(defmacro easy-kill-defun (name arglist &optional docstring &rest body)
+  "Like `defun' but NAME has property `easy-kill-exit' set to t.
+\n\n(fn NAME ARGLIST &optional DOCSTRING DECL &rest BODY)"
+  (declare (doc-string 3) (indent defun))
+  `(progn (put ',name 'easy-kill-exit t)
+          (defun ,name ,arglist ,docstring ,@body)))
+
 (defun easy-kill-echo (format-string &rest args)
   "Same as `message' except not writing to *Messages* buffer.
 Do nothing if `easy-kill-inhibit-message' is non-nil."
   (unless easy-kill-inhibit-message
     (let (message-log-max)
-      (apply 'message format-string args))))
+      (apply #'message format-string args))))
 
 (defun easy-kill-trim (s &optional how)
   (let ((wchars "[ \t\n\r\f\v]*"))
@@ -179,11 +163,10 @@ The value is the function's symbol if non-nil."
     (set-keymap-parent map easy-kill-base-map)
     (when easy-kill-unhighlight-key
       (with-demoted-errors "easy-kill-unhighlight-key: %S"
-        (define-key map easy-kill-unhighlight-key 'easy-kill-unhighlight)))
-    (mapc (lambda (c)
-            ;; (define-key map (vector meta-prefix-char c) 'easy-kill-select)
-            (define-key map (char-to-string c) 'easy-kill-thing))
-          (mapcar 'car easy-kill-alist))
+        (define-key map easy-kill-unhighlight-key #'easy-kill-unhighlight)))
+    (dolist (c easy-kill-alist)
+      ;; (define-key map (vector meta-prefix-char (car c)) #'easy-kill-select)
+      (define-key map (char-to-string (car c)) #'easy-kill-thing))
     map))
 
 (defun easy-kill--fmt (x y &optional z)
@@ -212,11 +195,8 @@ The value is the function's symbol if non-nil."
     (princ "\n")
     (princ (easy-kill--fmt "---" "-----" "---------"))
     (princ "\n\n")
-    (princ (mapconcat (lambda (x) (pcase x
-                                    (`(,c ,thing ,sep)
-                                     (easy-kill--fmt c thing sep))
-                                    ((or `(,c ,thing) `(,c . ,thing))
-                                     (easy-kill--fmt c thing))))
+    (princ (mapconcat (pcase-lambda (`(,c ,thing ,sep))
+                        (easy-kill--fmt c thing sep))
                       easy-kill-alist "\n"))
     (princ "\n\n")
     (princ (substitute-command-keys "\\{easy-kill-base-map}"))))
@@ -227,14 +207,9 @@ The value is the function's symbol if non-nil."
   (cons (overlay-start easy-kill-candidate)
         (overlay-end easy-kill-candidate)))
 
-;;; Note: gv-define-setter not available in 24.1 and 24.2
-;; (gv-define-setter easy-kill--bounds (val)
-;;   (macroexp-let2 macroexp-copyable-p v val
-;;     `(move-overlay easy-kill-candidate (car ,v) (cdr ,v))))
-
-(defsetf easy-kill--bounds () (v)
-  `(let ((tmp ,v))
-     (move-overlay easy-kill-candidate (car tmp) (cdr tmp))))
+(gv-define-setter easy-kill--bounds (val)
+  (macroexp-let2 macroexp-copyable-p v val
+    `(move-overlay easy-kill-candidate (car ,v) (cdr ,v))))
 
 (defmacro easy-kill-get (prop)
   "Get the value of the kill candidate's property PROP.
@@ -270,9 +245,6 @@ Use `setf' to change property value."
         (overlay-put o 'origin-indicator i)))
     (setq easy-kill-candidate o)
     (save-restriction
-      ;; Work around http://debbugs.gnu.org/15808; not needed in 24.4.
-      (narrow-to-region (max (point-min) (- (point) 1000))
-                        (min (point-max) (+ (point) 1000)))
       (let ((easy-kill-inhibit-message t))
         (cl-dolist (thing easy-kill-try-things)
           (easy-kill-thing thing n)
@@ -299,7 +271,7 @@ non-zero length, it is the string covered by the overlay.
 Otherwise, it is the value of the overlay's candidate property."
   (with-current-buffer (easy-kill-get buffer)
     (or (pcase (easy-kill-get bounds)
-          (`(,_x . ,_x) (easy-kill-get candidate))
+          (`(,x . ,x) (ignore x) (easy-kill-get candidate))
           (`(,beg . ,end) (filter-buffer-substring beg end)))
         "")))
 
@@ -377,6 +349,28 @@ candidate property instead."
   (interactive)
   (easy-kill-thing nil '+))
 
+(defun easy-kill-cycle (&optional thing)
+  "Cycle through things in `easy-kill-alist'.
+A thing is opted out of cycling if in `easy-kill-cycle-ignored'."
+  (interactive)
+  (let ((next (easy-kill-cycle-next (or thing (easy-kill-get thing))
+                                    (length easy-kill-cycle-ignored))))
+    (easy-kill-thing next)
+    (if (eq next (easy-kill-get thing))
+        (easy-kill-echo "%s" next)
+      ;; NEXT not killable continue cycle.
+      (easy-kill-cycle next))))
+
+(defun easy-kill-cycle-next (thing depth)
+  (let ((next (cl-second
+               (car (or (cl-loop for (head . tail) on easy-kill-alist
+                                 when (eq thing (cl-second head))
+                                 return tail)
+                        easy-kill-alist)))))
+    (cond ((not (memq next easy-kill-cycle-ignored)) next)
+          ((> depth 0) (easy-kill-cycle-next next (1- depth)))
+          (t (user-error "Nothing to cycle")))))
+
 (defun easy-kill-digit-argument (n)
   "Expand selection by N number of things.
 If N is 0 shrink the selection to the initial size before any
@@ -411,11 +405,19 @@ checked."
 
 (defun easy-kill-bounds-of-thing-at-point (thing)
   "Easy Kill wrapper for `bounds-of-thing-at-point'."
-  (pcase (easy-kill-thing-handler
-          (format "easy-kill-bounds-of-%s-at-point" thing)
-          major-mode)
-    ((and (pred functionp) fn) (funcall fn))
-    (_ (bounds-of-thing-at-point thing))))
+  ;; Work around a bug (fixed in 25.1, commit: 7a94f28a) in
+  ;; `thing-at-point-bounds-of-url-at-point' that could return a
+  ;; boundary not containing current point.
+  (cl-flet ((chk (bound)
+              (pcase-let ((`(,b . ,e) bound))
+                (and b e
+                     (<= b (point)) (<= (point) e)
+                     (cons b e)))))
+    (pcase (easy-kill-thing-handler
+            (format "easy-kill-bounds-of-%s-at-point" thing)
+            major-mode)
+      ((and (pred functionp) fn) (chk (funcall fn)))
+      (_ (chk (bounds-of-thing-at-point thing))))))
 
 (defun easy-kill-thing-forward-1 (thing &optional n)
   "Easy Kill wrapper for `forward-thing'."
@@ -441,7 +443,7 @@ checked."
                         (`1 end))))
            (new-front (save-excursion
                         (goto-char front)
-                        (with-demoted-errors
+                        (with-demoted-errors "%S"
                           (dotimes (_ (abs n))
                             (easy-kill-thing-forward-1 thing step)))
                         (point))))
@@ -454,9 +456,7 @@ checked."
 (defun easy-kill-thing (&optional thing n inhibit-handler)
   ;; N can be -, + and digits
   (interactive
-   (list (pcase (assq last-command-event easy-kill-alist)
-           (`(,_ ,th . ,_) th)
-           (`(,_ . ,th) th))
+   (list (cl-second (assq last-command-event easy-kill-alist))
          (prefix-numeric-value current-prefix-arg)))
   (let* ((thing (or thing (easy-kill-get thing)))
          (n (or n 1))
@@ -483,8 +483,7 @@ checked."
     (when (easy-kill-get mark)
       (easy-kill-adjust-candidate (easy-kill-get thing)))))
 
-(put 'easy-kill-abort 'easy-kill-exit t)
-(defun easy-kill-abort ()
+(easy-kill-defun easy-kill-abort ()
   (interactive)
   (when (easy-kill-get mark)
     ;; The after-string may interfere with `goto-char'.
@@ -493,20 +492,17 @@ checked."
     (setq deactivate-mark t))
   (ding))
 
-(put 'easy-kill-region 'easy-kill-exit t)
-(defun easy-kill-region ()
+(easy-kill-defun easy-kill-region ()
   "Kill current selection and exit."
   (interactive "*")
   (pcase (easy-kill-get bounds)
-    (`(,_x . ,_x) (easy-kill-echo "Empty region"))
+    (`(,x . ,x) (ignore x) (easy-kill-echo "Empty region"))
     (`(,beg . ,end) (kill-region beg end))))
 
-(put 'easy-kill-mark-region 'easy-kill-exit t)
-(defun easy-kill-mark-region ()
+(easy-kill-defun easy-kill-mark-region ()
   (interactive)
   (pcase (easy-kill-get bounds)
-    (`(,_x . ,_x)
-     (easy-kill-echo "Empty region"))
+    (`(,x . ,x) (ignore x) (easy-kill-echo "Empty region"))
     (`(,beg . ,end)
      (pcase (if (eq (easy-kill-get mark) 'end)
                 (list end beg) (list beg end))
@@ -517,13 +513,13 @@ checked."
 
 (defun easy-kill-exchange-point-and-mark ()
   (interactive)
-  (exchange-point-and-mark)
-  (setf (easy-kill-get mark)
-        (if (eq (point) (easy-kill-get start))
-            'end 'start)))
+  (when (easy-kill-get mark)
+    (exchange-point-and-mark)
+    (setf (easy-kill-get mark)
+          (if (eq (point) (easy-kill-get start))
+              'end 'start))))
 
-(put 'easy-kill-append 'easy-kill-exit t)
-(defun easy-kill-append ()
+(easy-kill-defun easy-kill-append ()
   (interactive)
   (setf (easy-kill-get append) t)
   (when (easy-kill-save-candidate)
@@ -531,10 +527,15 @@ checked."
     (setq deactivate-mark t)
     (easy-kill-echo "Appended")))
 
-(put 'easy-kill-unhighlight 'easy-kill-exit t)
-(defun easy-kill-unhighlight ()
+(easy-kill-defun easy-kill-delete-region ()
   (interactive)
-  (easy-kill-echo "`%s' copied" (easy-kill-get thing)))
+  (pcase (easy-kill-get bounds)
+    (`(,beg . ,end) (delete-region beg end))))
+
+(easy-kill-defun easy-kill-unhighlight ()
+  (interactive)
+  (and (easy-kill-save-candidate)
+       (easy-kill-echo "`%s' copied" (easy-kill-get thing))))
 
 (defun easy-kill-exit-p (cmd)
   (and (symbolp cmd) (get cmd 'easy-kill-exit)))
@@ -571,6 +572,7 @@ Temporally activate additional key bindings as follows:
   @       => append selection to previous kill;
   ?       => help;
   C-w     => kill selection;
+  SPC     => cycle through things in `easy-kill-alist';
   C-SPC   => turn selection into an active region;
   C-g     => abort;
   others  => save selection and exit."
@@ -587,7 +589,7 @@ Temporally activate additional key bindings as follows:
     (easy-kill-activate-keymap)))
 
 ;;;###autoload
-(defalias 'easy-mark-sexp 'easy-mark
+(defalias 'easy-mark-sexp #'easy-mark
   "Use `easy-mark' instead. The alias may be removed in future.")
 
 ;;;###autoload
@@ -783,8 +785,8 @@ inspected."
     (n (ignore-errors
          (dotimes (_ (abs n))
            (pcase (list (point) (easy-kill-bounds-of-thing-at-point 'list))
-             (`(,_beg (,_beg . ,_)) (org-up-element))
-             (`(,_ (,beg . ,_))     (goto-char beg)))))
+             (`(,beg (,beg . ,_)) (ignore beg) (org-up-element))
+             (`(,_ (,beg . ,_))   (goto-char beg)))))
        (when (cl-plusp n)
          (goto-char (cdr (easy-kill-bounds-of-thing-at-point 'list)))))))
 
